      subroutine cmpt_ort(fft_klh, hkl_max, fft_ann, ctf_val,
     $     abs_ctf, disef, pixels, theta, phi, omega, x, y,
     $     mag, score, orient_flag, origin_flag)

      implicit none

      include '../include/sizes.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'

      integer, intent(in) :: hkl_max, fft_ann
      complex, intent(in) :: fft_klh(-fft_ann:fft_ann, 
     $     -fft_ann:fft_ann, 0:fft_ann)
      real, intent(in) :: ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in) :: abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in) :: disef(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in) :: pixels(npix_row, npix_col)

      real, intent(inout) :: theta, phi, omega, x, y
      real, intent(in)    :: mag
      real, intent(out)   :: score
      character(len=1), intent(out) :: orient_flag, origin_flag

      real theta_save, phi_save, omega_save

      integer i, j
      integer :: ishift, maxshift = 10
      integer theta_step, phi_step, omega_step
      integer centeru, centerv
      integer best_step(3), min_cen(2), minu, minv
      integer alloc_stat

      real :: zero = 0.0
      real :: r4max = huge(r4max)
      real center(2), imgmagfactor, theta_test, phi_test, omega_test,
     $     num_annul,value,min_value, min_annul, du, dv,
     $     rot_mat(3,3), distcen(-Ncenter:Ncenter,-Ncenter:Ncenter),
     $     dist(-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle),
     $     delta_theta, delta_phi, delta_omega
      
      real, allocatable :: rffta_omega(:), unit_array(:,:)
      complex, allocatable :: fc_2d(:,:), fc_2d_intp(:,:)
      complex, allocatable :: fc_2d_bak(:,:)
      
      logical stable_orient, stable_origin, stable_hand

C     Avoid calculations on ptles with terrible origins
      if(  x < npix_row*0.1 .or. x > npix_row*0.9 .or.
     $     y < npix_col*0.1 .or. y > npix_col*0.9) then
         score = 9999.999
         orient_flag = 'X'
         origin_flag = 'X'
         return
      endif

      allocate(rffta_omega(nrffta_omega),
     +     fc_2d(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     fc_2d_intp(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     fc_2d_bak(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     unit_array(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     stat = alloc_stat)
      
C     Initialize unit array using Fortran 90 array syntax
      unit_array = 1.0

      if(alloc_stat > 0) then
         call error_stop('PO2R', 'cmpt_ort', ' Memory alloc failed')
      endif

C     Initialize the array rffta_omega that is used by FFT library
      call vrffti(nrffta_col, rffta_omega)
      
      num_annul = real(nrffta_row)*angstrom_per_pix/res_max 
      min_annul = real(nrffta_row)*angstrom_per_pix/res_min 
      if (num_annul >= nrffta_row/2.) 
     +     call error_stop('PO2R','cmpt_ort', 
     +     'Resolution is too high, smaller than 2*angstrom_per_pix')
      
      theta_save = theta
      phi_save   = phi
      omega_save = omega
      center(1)    = x
      center(2)    = y
      imgmagfactor = mag
            
C     adjust the imgmagfactor according to its own pixel_size
      imgmagfactor = imgmagfactor * PIXSIZ_OUT / angstrom_per_pix
            
      call fft_2dfft(pixels, npix_row, npix_col, center, nrffta_omega,
     $     rffta_omega, nrffta_row, nrffta_col, nfc_2d_row, nfc_2d_col,
     $     fc_2d )

C     Need to apply tempfac to images
      call maptempfac(fc_2d, 0, nfc_2d_row, -nfc_2d_col,
     $     nfc_2d_col, 1, 1, tempfac, pixsiz_out, units)
      
      du = center(1) - real(int(center(1)))
      dv = center(2) - real(int(center(2)))
      call shift_fft(fc_2d, nfc_2d_row, nfc_2d_col, du, dv)
      
C     backup fc_2d
      fc_2d_bak = fc_2d         ! f90 array syntax
      fc_2d_intp = 0.           ! f90 array syntax
      
      min_value = R4MAX
      
C     ----------- Refine particle orientation -----------
C     
C     Search a region of orientation space, centered on the current
C     orientation, for a new model projection that has the best
C     agreement (lowest score) with the particle image. The number of
C     steps taken in each direction for each angle is nangle. If the new
C     optimal orientation is more than nangle/2 steps from the starting
C     point in any direction, perform another search centered on this
C     new orientation. To minimize the number of comparisons that need
C     to be calculated, the array DIST containing the scores is
C     recentered so that previously calculated scores can be reused. The
C     outer loop over ishift sets a limit on the number of times that
C     search can be restarted.
C
C     Note that in this convention, the angle omega has to be set to
C     (360-omega-90), or equivalently, -(omega+90).

C     Do a very minimal local search first to determine whether or not
C     to perform a more extensive search
            
      dist = R4MAX
      best_step(1) = 0
      best_step(2) = 0
      best_step(3) = 0
      
      if(quick_search == 1 .and. nangle > 0) then
         do omega_step = -1, 1
            do phi_step = -1, 1
               do theta_step = -1, 1
                  
                  delta_theta = dangle * theta_step
                  delta_phi   = dangle * phi_step
                  delta_omega = dangle * omega_step
                  
                  theta_test = delta_theta + theta
                  phi_test   = delta_phi   + phi
                  omega_test = delta_omega + omega
                  
                  call crowther_to_matrix(theta_test, -phi_test,
     $                 -omega_test, rot_mat)
                  call interpl_3d(fft_klh, hkl_max, min_annul, 
     $                 num_annul, rot_mat, fc_2d_intp, imgmagfactor,
     +                 fft_ann)
                  
                  fc_2d_intp = fc_2d_intp * ctf_val
                  
                  call imgcomport(fc_2d_intp, hkl_max, fc_2d, disef,
     $                 abs_ctf, min_annul, num_annul, value, funcmode)
                  dist(theta_step, phi_step, omega_step)=value
                  
                  if (value < min_value) then
                     min_value = value
                     best_step(1) = theta_step
                     best_step(2) = phi_step
                     best_step(3) = omega_step
                  endif
               enddo
            enddo
         enddo
      endif
      
      if(nangle > 0 .and.
     $     (quick_search .ne. 1 .or. best_step(1) .ne. 0 .or.
     $     best_step(2)  .ne. 0 .or. best_step(3) .ne. 0)) then
         
         min_value = R4MAX
         dist = R4MAX
         
         do ishift = 1, maxshift
            
            best_step(1) = 0
            best_step(2) = 0
            best_step(3) = 0
            
            do omega_step = -Nangle, Nangle  
               do phi_step = -Nangle, Nangle  
                  do theta_step = -Nangle, Nangle  
                     
                     delta_theta = dangle * theta_step
                     delta_phi   = dangle * phi_step
                     delta_omega = dangle * omega_step
                     
                     theta_test = delta_theta + theta
                     phi_test   = delta_phi   + phi
                     omega_test = delta_omega + omega
                     
                     if (dist(theta_step,phi_step,omega_step) >= 
     +                    R4MAX-1) then
                        call crowther_to_matrix(theta_test, 
     $                       -phi_test ,-omega_test, rot_mat)
                        call interpl_3d(fft_klh, hkl_max, min_annul,
     $                       num_annul, rot_mat, fc_2d_intp,
     $                       imgmagfactor, fft_ann) 
                        
                        fc_2d_intp = fc_2d_intp * ctf_val
                        
                        call imgcomport(fc_2d_intp, hkl_max, 
     +                       fc_2d, disef, abs_ctf, min_annul, 
     +                       num_annul, value, funcmode)
                        dist(theta_step,phi_step,omega_step)=value
                        
                        if (value < min_value) then
                           min_value = value
                           best_step(1) = theta_step
                           best_step(2) = phi_step
                           best_step(3) = omega_step
                        endif
                     endif
                  enddo
               enddo
            enddo
            
            theta = theta + best_step(1) * dangle
            phi   = phi   + best_step(2) * dangle
            omega = omega + best_step(3) * dangle
            
C     If the new optimal orientation is sufficiently far from the
C     starting orientation, recenter the array DIST and start a new
C     search. Otherwise, finished with orientation refinement.
            
            if ( abs(real(best_step(1))) > Nangle*0.5 .or. 
     +           abs(real(best_step(2))) > Nangle*0.5 .or.
     +           abs(real(best_step(3))) > Nangle*0.5 ) then
               call recenter_array_3d(dist,Nangle,best_step,R4MAX)
            else
               exit
            endif
         enddo
      endif
      
      if(theta_save == theta .and. phi_save == phi .and.
     $     omega_save == omega) then
         stable_orient = .true.
      else
         stable_orient = .false.
      endif
      
C     ----------- Check handedness of orientation -----------
C     
C     Compare current orientation with the corresponding orientation
C     with the opposite handedness (theta, -phi, omega+180). Switch
C     handedness of the opposite handed orientation has a better match
C     with the model.
      
      theta_save = theta
      phi_save   = phi
      omega_save = omega
      
C     Calculate score for current orientation
      call crowther_to_matrix(theta, -phi, -omega, rot_mat)
      call interpl_3d(fft_klh, hkl_max, min_annul, num_annul, rot_mat,
     $     fc_2d_intp, imgmagfactor, fft_ann) 
      fc_2d_intp = fc_2d_intp * ctf_val
      call imgcomport(fc_2d_intp, hkl_max, fc_2d, disef, abs_ctf,
     $     min_annul, num_annul, min_value, funcmode)

      if(handtest == 1) then
C        Calculate score for hand-flipped orientation
         call crowther_to_matrix(theta, phi, -omega+180, rot_mat)
         call interpl_3d(fft_klh, hkl_max, min_annul, num_annul, 
     $        rot_mat, fc_2d_intp, imgmagfactor, fft_ann) 
         fc_2d_intp = fc_2d_intp * ctf_val
         call imgcomport(fc_2d_intp, hkl_max, fc_2d, disef, abs_ctf,
     $        min_annul, num_annul, value, funcmode)
         
         if (value < min_value) then
            min_value = value
            phi = - phi
            omega = omega + 180.0
         endif
         
         if(theta_save == theta .and. phi_save == phi .and. 
     $        omega_save == omega) then
            stable_hand = .true.
         else
            stable_hand = .false.
         endif
      else
         stable_hand = .true.
      endif
      
C     ----------- Return orientation to asymmetric unit -----------
C     
      if ( symm_code > 1) then
         call to_asym_unit(symm_code, theta, phi, omega)
      endif
      
C     ----------- Refine particle origin -----------
C     
C     Search a region of origin space, centered on the current origin,
C     for a new model projection that has the best agreement (lowest
C     score) with the particle image. The number of steps taken in each
C     direction for each coordinate is ncenter. If the new optimal
C     origin is more than ncenter/2 steps from the starting point in any
C     direction, perform another search centered on this new origin. To
C     minimize the number of comparisons that need to be calculated, the
C     array DISTCEN containing the scores is recentered so that
C     previously calculated scores can be reused. The outer loop over
C     ishift sets a limit on the number of times that search can be
C     restarted.
      
      call crowther_to_matrix(theta, -phi, -omega, rot_mat)
      call interpl_3d(fft_klh, hkl_max,
     +     min_annul, num_annul,
     +     rot_mat, fc_2d_intp, imgmagfactor, fft_ann) 
      
      fc_2d_intp = fc_2d_intp * ctf_val
      
      distcen = R4MAX           ! f90 array syntax
      
      min_value = R4MAX
      minu      = 0
      minv      = 0
      
      do ishift = 1, maxshift
         
         min_cen(1) = 0
         min_cen(2) = 0
         
         do centerv = -Ncenter, Ncenter 
            dv = Dcenter * (centerv + minv)
            do centeru = -Ncenter, Ncenter 
               if (distcen(centeru,centerv) >= R4MAX-1) then
                  du = Dcenter * (centeru + minu)
                  fc_2d = fc_2d_bak
                  call shift_fft(fc_2d,
     +                 nfc_2d_row, nfc_2d_col, du, dv)
                  
                  call imgcomport(fc_2d_intp, hkl_max, 
     +                 fc_2d, unit_array, abs_ctf, min_annul,
     +                 num_annul, value, 1)
                  
                  distcen(centeru,centerv) = value
                  if (value < min_value) then
                     min_value = value
                     min_cen(1) = centeru
                     min_cen(2) = centerv
                  endif
               endif
            enddo
         enddo
         
C     If the new optimal origin is sufficiently far from the starting
C     origin, recenter the array DISTCEN and start a new
C     search. Otherwise, finished with origin refinement.
         
         if ( abs(real(min_cen(1))) > Ncenter*0.5 .or. 
     +        abs(real(min_cen(2))) > Ncenter*0.5 ) then
            minu = minu + min_cen(1) 
            minv = minv + min_cen(2)
            call recenter_array_2d(distcen,Ncenter,min_cen,R4MAX)
         else
            exit
         endif
         
      enddo
      
      du = (minu+min_cen(1))*Dcenter
      dv = (minv+min_cen(2))*Dcenter
      
      if(du == 0 .and. dv == 0) then
         stable_origin = .true.
      else
         stable_origin = .false.
      endif
      
      orient_flag = 'S'
      origin_flag = 'S'
      
      if(stable_orient .and. stable_hand .and. .not.stable_origin) then
         origin_flag = 'N'
      endif

      if(.not.stable_orient .and. stable_origin) then
         orient_flag = 'N'
      endif

      if(.not. stable_orient .and. .not.stable_origin) then
         orient_flag = 'N'
         origin_flag = 'N'
      endif

      if(stable_orient .and. .not.stable_hand .and. stable_origin) then
         orient_flag = 'F'
      endif

      if(stable_orient .and. .not.stable_hand .and. 
     $     .not.stable_origin) then
         orient_flag = 'F'
         origin_flag = 'N'
      endif

      fc_2d = fc_2d_bak
      call shift_fft(fc_2d, nfc_2d_row, nfc_2d_col, du, dv)
      
C     compute the CC value for the DFT and best CUT
      call imgcomport(fc_2d_intp, hkl_max, fc_2d, disef, abs_ctf,
     $     min_annul, num_annul, min_value, funcmode)
            
      x = x + du
      y = y + dv
      score = min_value

      deallocate(rffta_omega, fc_2d, fc_2d_intp, fc_2d_bak, unit_array)
      
      return
      end subroutine cmpt_ort
