#include <math.h>
#include <stdlib.h>
#include "../include/common.h"
#include "../include/libPIF.h"

#define WRITEFFT   0  /* 1 = will write out modified fft */
#define WRITEIMAGE 5

#define PI         3.141592654

/*#define SPHERICALABBR 2 */ /* 2 mm */
/*#define AMPCONTRAST  .07*/  /* 7% */
/*#define PHASECONTRAST .93 */


/* prototypes */

float calc_ctf( float waveLength, float defocus_major, float kbold,
	       float thisSphericalAbbr, float ampContrast,
	       float phi, float phi_major, float defocus_minor );

/* end prototypes */

/*****************************************************
 *    APPLY CTF TO FFT OF MODEL IF CTFMODE > 0      *
 *     CTFMODE <= 0: DON'T APPLY CTF                *
 *             >= 1: APPLY FULL CTF                 *
 ****************************************************
*/
void apply_ctf_rtn(int *ctfmode, float *fft_float_buffer, int *ncol, 
		   int *nrow, int *idim1, int *idim2, int *view, 
		   float *volts, float *pixelSize,
		   float *ampContrast_in,
		   float *defocus_in, float *defocus_minor_in,
		   float *input_phi, float *sphericalAbbr
		   )
{ /* apply_ctf_rtn */

  float defocus, thisSphericalAbbr, thisPixelSize,  waveLength;
  float ampContrast, radius, kbold;
  float ctf_atpixel, phi, phi_major, defocus_minor;
  float Amplitude, AmplitudeSquared, Phase;
  float d;
  int   ii, jj, junk;
  int   thispixel;
  int   fftSize;
  FILE *fp;
  /*
   *printf("inputs:\nctfmode %d\tncol %d\tnrow %d\n",*ctfmode,*ncol,*nrow);
   *printf("idim1 %d idim2 %d view %d volts %f\n",*idim1,*idim2,*view,*volts);
   *printf("pixel %f ampcontrast %f\n",*pixelSize,*ampContrast_in);
   *printf("defocus %f minor %f angle %f Cs %f\n",
   *       *defocus_in,*defocus_minor_in,
   *       *input_phi,*sphericalAbbr);
   */

  ampContrast = *ampContrast_in;
  if ( *idim1 != *idim2)
    {
      printf("For CTF correction, idim1 must = idim2\n");
      exit(1);
    }
  fftSize = *idim1;

  /* input phi is in degrees, want radians */
  phi_major = (*input_phi / 180.0) * PI;

  defocus = *defocus_in;
  defocus_minor = *defocus_minor_in;
  if (defocus < 0.0)
    printf("Warning: underfocus is positive\n");
  if (defocus_minor == 0.0)
    defocus_minor = defocus;

 
  /* lets do everything in nm for sanity */
  defocus *= 1000;        /* defocus entered in microns */
  defocus_minor *= 1000;  
  thisSphericalAbbr = *sphericalAbbr * 1000000.0;
  thisPixelSize = *pixelSize / 10.0;
  waveLength = 1.23 / sqrt(*volts + (1.0e-6 * *volts * *volts));

  /*  printf("Using %f kV\n", (volts / 1000) ); */
  /*  printf("Wave Length is %f nm\n", waveLength); */
  
  /*  printf("for view %d,\n", *view);
   * printf("defocus (major axis) = %f, defocus (minor axis) = %f in nm\n",
   *	 defocus, defocus_minor);
   * printf("angle to major axis = %f, pixel size = %f nm\n",
   *	 phi_major, thisPixelSize);
   *printf("spherical abbration = %f\n",  thisSphericalAbbr);
   */

  /* jj == IDIM 1 in the fortran */
  /* ii == IDIM 2 in the fortran */

  for(jj=0; jj < fftSize; jj++) 
    { /* for( jj */
      for(ii=0; ii < (fftSize); ii++) 
	{ /* for ( ii */

	  /* calculate distance from center in pixels  */
	  
	  /* quadrants organized as 4 3                          * 
	   *                        1 2                          *
	   * with jj,ii = 0,0 located on the bottom left corner  *
	   * and  jj,ii = fftSize-1,fftSize-1 located            *
	   * on the top right corner                             */

	  thispixel = (int)(( ii * ( fftSize * 2 )) + (2 * jj));
	  
	  /* in fft space, the origin is at the bottom left corner */
	  if ( jj < (fftSize / 2) ) 
	    {
	      if ( ii < (fftSize / 2) )
		{ /* quadrant 1 */
		  radius = sqrt((double) (ii * ii) + (jj * jj) );
		  /*if (jj == 23)
		   * {
		   *printf("\nii %d jj%d radius%f\n", ii, jj, radius);
		   *printf("calc = %f\n", sqrt( (ii * ii) + (jj * jj) ));
		   *}
		   */

	  /* for astigmatism */
		  phi =  atan2( (float) ii, (float) jj );
		} /* quadrant 1 */
	      else
		{ /* quadrant 4 */
		  radius = sqrt((double) ((fftSize-ii)*(fftSize-ii))+(jj*jj));
		  phi = atan2( (float)(fftSize-ii), (float) jj);
		} /* quadrant 4 */
	    }
	  else /* either quad. 2 or 3 */
	    {
	      if ( ii < fftSize / 2)
		{ /* quadrant 2 */
		  radius = sqrt((double)(ii*ii)+((fftSize -jj)*(fftSize -jj)));
		  phi = atan2( (float) ii, (float) (fftSize - jj) );
		} /* quadrant 2 */
	      else
		{ /* quadrant 3 */
		  radius = sqrt((double) ((fftSize-ii)*(fftSize-ii)) +
				((fftSize - jj)*(fftSize - jj)) );
		  phi = atan2( (float)(fftSize-ii), (float)(fftSize-jj));
		} /* quadrant 3 */
	    } /* end either quad. 2 or 3 */
	  
	  /* in fft space, 1 pixel is actually one inverse pixel *
	   * so, one pixel = (1/ (fftSize * pixel size)) in inverse distance */
	  
	  kbold = radius / (fftSize * thisPixelSize);
	  /* so, kbold is in inverse nm */
	  /*if (jj==23)
	   *{
	   *printf("kbold %f phi %f ii%d jj%d\n", kbold,phi,ii,jj);
	   *printf("radius %f, fftSize %d\n", radius, fftSize);
	   *}
	   */
	  
	  ctf_atpixel = calc_ctf( waveLength, defocus, kbold, 
				 thisSphericalAbbr, ampContrast,
				 phi, phi_major, defocus_minor);



	  /* first, put in amplitue and phases for my understanding */
	  /* and for later amplitude correction */


	  AmplitudeSquared = ( fft_float_buffer[thispixel] *
			      fft_float_buffer[thispixel] ) + 
				( fft_float_buffer[thispixel + 1] * 
				 fft_float_buffer[thispixel + 1]);
	  Amplitude = sqrt( AmplitudeSquared ); 
	  /* the atan2 is atan2(y/x) */
	  
	  /* for flipping phases, if ctf is negative do nothing *
	   *                      if ctf is positive flip       */
	  
	  Phase = atan2( fft_float_buffer[thispixel+1], 
			fft_float_buffer[thispixel]); 

	  if(*ctfmode >= 1)
	    {/* if apply ctf == yes */	  
	      /* now, apply ctf */
	      Amplitude *= (ctf_atpixel * (-1.0));
	    }/* if apply ctf == yes */	  

	  /* put back into a +ib form and into buffer */
	  /* a = (amp)cos(phase), b = (amp)sin(phase) */
	  
	  fft_float_buffer[thispixel] = Amplitude * cos(Phase);
	  fft_float_buffer[thispixel + 1] = Amplitude * sin(Phase);

	} /* for( ii   */
    } /* for( jj  */

  if (*view == WRITEIMAGE)
    { /* view == WRITEIMAGE */
      if ( WRITEFFT )
	{ /* WRITEFFT */
	  fp = fopen("em3dr_modified_fft.dat", "w");
	  fwrite( fft_float_buffer, sizeof(float), 2 * fftSize * fftSize,
		 fp);
	} /* WRITEFFT */
    } /* view == WRITEIMAGE */

} /* apply_ctf_rtn */


/* Define this for portability */
void apply_ctf_rtn_(int *ctfmode, float *fft_float_buffer, int *ncol, 
		    int *nrow, int *idim1, int *idim2, int *view, 
		    float *volts, float *pixelSize,
		    float *ampContrast_in,
		    float *defocus_in, float *defocus_minor_in,
		    float *input_phi, float *sphericalAbbr
		    )
{/* APPLY_CTF_RTN */
  apply_ctf_rtn(ctfmode, fft_float_buffer, ncol, 
		nrow, idim1, idim2, view, 
		volts, pixelSize,
		ampContrast_in,
		defocus_in, defocus_minor_in,
		input_phi, sphericalAbbr
		   );
}/* APPLY_CTF_RTN */





/*****************************************************
 *    APPLY CTF CORRECTION TO FFT IF CTFMODE > 0    *
 *     CTFMODE = 0: DON'T APPLY ANY CORRECTION.     *
 *            = 1: FULL CTF CORRECTION              *
 *             = 2: PHASE FLIP ONLY                 *
 *             = 3: AMPLITUDE CORRECT ONLY          *
 ****************************************************
*/

void calc_ctf_rtn(int *ctfmode, float *fft_float_buffer, int *ncol, 
		   int *nrow, int *idim1, int *idim2, int *view, 
		   float *volts, float *pixelSize, float *wiener,
		   float *ampContrast_in, float *tFactor,
		   float *defocus_in, float *defocus_minor_in,
		   float *input_phi, float *sphericalAbbr
		   )
{ /* calc_ctf_rtn */

  float defocus, thisSphericalAbbr, thisPixelSize,  waveLength;
  float ampContrast, radius, kbold;
  float ctf_atpixel, phi, phi_major, defocus_minor;
  float Amplitude, AmplitudeSquared, Phase;
  float S, H;
  float d, Temp, TempFactor;
  int   ii, jj;
  int   thispixel;
  int   fftSize;
  FILE *fp;
  
  /*
   *printf("inputs:\nctfmode %d\tncol %d\tnrow %d\n",*ctfmode,*ncol,*nrow);
   *printf("idim1 %d idim2 %d view %d volts %f\n",*idim1,*idim2,*view,*volts);
   *printf("pixel %f wiener %f ampcontrast %f, 
   *tfactor %f\n",*pixelSize,*wiener,
   *       *ampContrast_in, *tFactor);
   *printf("defocus %f minor %f angle %f Cs %f\n",
   *       *defocus_in,*defocus_minor_in,
   *       *input_phi,*sphericalAbbr);
   */
  
  TempFactor = *tFactor;
  ampContrast = *ampContrast_in;
  if ( *idim1 != *idim2)
    {
      printf("For CTF correction, idim1 must = idim2\n");
      exit(1);
    }
  fftSize = *idim1;

  /* input phi is in degrees, want radians */
  phi_major = (*input_phi / 180.0) * PI;

  defocus = *defocus_in;
  defocus_minor = *defocus_minor_in;
  if (defocus < 0.0)
    printf("Warning: underfocus is positive\n");
  if (defocus_minor == 0.0)
    defocus_minor = defocus;

  /*  printf("Defocus is %f nm\n", defocus); */
  /*  printf("Spherical Abbration is %d mm\n", SPHERICALABBR); */
  /*  printf("Aplitude contrast is %f\n", AMPCONTRAST); */
  /*  printf("Pixel size is %f angstroms/pixel\n", PIXELSIZE); */
 
  /* lets do everything in nm for sanity */
  defocus *= 1000;        /* defocus entered in microns */
  defocus_minor *= 1000;  
  thisSphericalAbbr = *sphericalAbbr * 1000000.0;
  thisPixelSize = *pixelSize / 10.0;
  TempFactor = TempFactor / 100.0; 
  waveLength = 1.23 / sqrt(*volts + (1.0e-6 * *volts * *volts));

  /*  printf("Using %f kV\n", (volts / 1000) ); */
  /*  printf("Wave Length is %f nm\n", waveLength); */
  
/*  printf("for view %d,\n", *view);
 *  printf("defocus (major axis) = %f, defocus (minor axis) = %f in nm\n",
 *	 defocus, defocus_minor);
 *  printf("angle to major axis = %f, pixel size = %f nm\n",
 *	 phi_major, thisPixelSize);
 */ 

  /* jj == IDIM 1 in the fortran */
  /* ii == IDIM 2 in the fortran */

  for(jj=0; jj < fftSize; jj++) 
    { /* for( jj */
      for(ii=0; ii < (fftSize); ii++) 
	{ /* for ( ii */

	  /* calculate distance from center in pixels  */
	  
	  /* quadrants organized as 4 3                          * 
	   *                        1 2                          *
	   * with jj,ii = 0,0 located on the bottom left corner  *
	   * and  jj,ii = fftSize-1,fftSize-1 located            *
	   * on the top right corner                             */

	  thispixel = (int)(( ii * ( fftSize * 2 )) + (2 * jj));
	  
	  /* in fft space, the origin is at the bottom left corner */
	  if ( jj < (fftSize / 2) ) 
	    {
	      if ( ii < (fftSize / 2) )
		{ /* quadrant 1 */
		  radius = sqrt((double)((ii * ii) + (jj * jj)) );
		  /* for astigmatism */
		  phi =  atan2( (float) ii, (float) jj );
		} /* quadrant 1 */
	      else
		{ /* quadrant 4 */
		  radius = sqrt( (double)((fftSize-ii)*(fftSize-ii)) + 
				(double)( jj * jj));
		  phi = atan2( (float)(fftSize-ii), (float) jj);
		} /* quadrant 4 */
	    }
	  else /* either quad. 2 or 3 */
	    {
	      if ( ii < fftSize / 2)
		{ /* quadrant 2 */
		  radius = sqrt((double) (ii*ii) + 
				(double)( (fftSize - jj)*(fftSize - jj) ) );
		  phi = atan2( (float) ii, (float) (fftSize - jj) );
		} /* quadrant 2 */
	      else
		{ /* quadrant 3 */
		  radius = sqrt( (double)((fftSize-ii)*(fftSize-ii)) +
				(double)((fftSize - jj)*(fftSize - jj)) );
		  phi = atan2( (float)(fftSize-ii), (float)(fftSize-jj));
		} /* quadrant 3 */
	    } /* end either quad. 2 or 3 */

	  
	  /* in fft space, 1 pixel is actually one inverse pixel *
	   * so, one pixel = (1/ (fftSize * pixel size)) in inverse distance */
	  
	  kbold = radius / (fftSize * thisPixelSize);
	  /* so, kbold is in inverse nm */
	  
	  ctf_atpixel = calc_ctf( waveLength, defocus, kbold, 
				 thisSphericalAbbr, ampContrast,
				 phi, phi_major, defocus_minor);

	  
	  /* first, put in amplitue and phases for my understanding */
	  /* and for later amplitude correction */


	  AmplitudeSquared = ( fft_float_buffer[thispixel] *
			      fft_float_buffer[thispixel] ) + 
				( fft_float_buffer[thispixel + 1] * 
				 fft_float_buffer[thispixel + 1]);
	  Amplitude = sqrt((double) AmplitudeSquared ); 
	  /* the atan2 is atan2(y/x) */
	  
	  /* for flipping phases, if ctf is negative do nothing *
	   *                      if ctf is positive flip       */
	  
	  Phase = atan2( fft_float_buffer[thispixel+1], 
			fft_float_buffer[thispixel]); 


	  if( (*ctfmode == 1) || (*ctfmode ==2) )
	    {/* if phasect == yes */	  
	      /* now, correct phases, amplitudes if desired/needed */
	      if (ctf_atpixel > 0.0) 
		{
		  Phase -= PI;
		} /* end if ctf > 0.0 */
	    }/* if phasect == yes */	  
	
	  /*to correct amplitudes, use modified filer   */
	  /* F = SI, where I == image                              *
	   *               F == "actual" function                  *
	   *               S == filter                             * 
	   * we have already flipped phases, so now use fabs(ctf)  *
	   *     H = fabs(ctf)                                     *
	   *     W = wiener filter                                 *
	   *   S = H / ( H*H + W )                                 */

	  if( (*ctfmode == 1) || (*ctfmode == 3) )
	    {/* if ampCorr = yes */
	      H = fabs(ctf_atpixel);
	      S = H / ( (H*H) + *wiener );
	      Amplitude *= S;
	    }/* if ampCorr = yes */

	  /* now, apply temperature factor   */
	  /* Fafter = T * Fobserved *
	   *   T = exp ( (B/4) * ((1/d)^2) ) *
	   * we will use both B and d (= 1/kbold) in nm  */
	  /* so T = exp ( (B/4) * (kbold)^2 )  */
	  
	  if(TempFactor > 0.0)
	    {/* if tempfactor > 0.0) */
	      /* note: conditional not stictly necessary */
	      Temp = exp( (TempFactor/4.0) * (kbold * kbold));
	      /*	  printf("temp = %f, kbold = %f\n", Temp, kbold); */
	      Amplitude *= Temp;
	    }/* if tempfactor > 0.0) */

	  /* put back into a +ib form and into buffer */
	  /* a = (amp)cos(phase), b = (amp)sin(phase) */
	  
	  fft_float_buffer[thispixel] = Amplitude * cos(Phase);
	  fft_float_buffer[thispixel + 1] = Amplitude * sin(Phase);

	} /* for( ii   */
    } /* for( jj  */

  if (*view == WRITEIMAGE)
    { /* view == WRITEIMAGE */
      if ( WRITEFFT )
	{ /* WRITEFFT */
	  fp = fopen("em3dr_modified_fft.dat", "w");
	  fwrite( fft_float_buffer, sizeof(float), 2 * fftSize * fftSize,
		 fp);
	} /* WRITEFFT */
    } /* view == WRITEIMAGE */

} /* calc_ctf_rtn */


/* Define this for portability */
void calc_ctf_rtn_(int *ctfmode, float *fft_float_buffer, int *ncol, 
		   int *nrow, int *idim1, int *idim2, int *view, 
		   float *volts, float *pixelSize, float *wiener,
		   float *ampContrast_in, float *tFactor,
		   float *defocus_in, float *defocus_minor_in,
		   float *input_phi, float *sphericalAbbr
		   )
{/* CALC_CTF_RTN */
  calc_ctf_rtn(ctfmode, fft_float_buffer, ncol, 
	       nrow, idim1, idim2, view, 
	       volts, pixelSize, wiener,
	       ampContrast_in, tFactor,
	       defocus_in, defocus_minor_in,
	       input_phi, sphericalAbbr
	       );
}/* CALC_CTF_RTN */





/***************************************************************/
/*  calc_ctf() - returns value of ctf for given parameters */

float calc_ctf( float waveLength, float defocus_major, float kbold,
	       float thisSphericalAbbr, float ampContrast,
	       float phi, float phi_major, float defocus_minor)
{ /* calc_ctf */
  
  float phaseContrast;
  float gammaA, gammaB, gamma, ctf, defocus;
  float phi_reference, mean_defocus, max_defocus_difference;

  phaseContrast = 1.0 - ampContrast;
  
  /* for astmatisim, input is defocus at major axis, = defocus_major *
   *                          defocus at minor axis, = defocus_minor *
   *                          angle to major axis    = phi_major     *
   *                          angle of thispixel     = phi           *
   * note that angle is assumed to start from the positive x axis    *
   * and increase to the counterclockwise direction                  */

  if ( defocus_minor != 0.0 )
    { /* determine actual defocus */
      phi_reference = phi_major - (PI / 4.0);
/* printf("phi_reference = %f, phi = %f\n", phi_reference, phi); */
      mean_defocus = (defocus_major + defocus_minor) / 2.0;
      max_defocus_difference = defocus_major - defocus_minor;
      defocus = mean_defocus + 
	( (max_defocus_difference / 2.0) * 
	 sin( 2.0 * ( phi - phi_reference)) );
/* printf("sin = %f\n", sin( 2.0 * ( phi - phi_reference)) ); */
    } /* determine actual defocus */
  else
    { /* no astigmation correction */
      defocus = defocus_major;
    } /* no astigmation correction */
/* printf("defocus_major %f, defocus_minor %f, defocus %f\n", defocus_major,
       defocus_minor, defocus); */

  /* let's calc ctf piece by piece *
   * gamma = 2Pi * (gammaA + gammaB) *
   * gammaA = -(1/2)(lambda)(deltaZ)(k^2) *
   * gammaB = (1/4)(lambda^3)(Cs)(k^4) *
   * CTF = (%phase)sin(gamma) - (%amplitude)cos(gamma) */
  
  gammaA = (-1.0) * (1.0/2.0) * waveLength * defocus * kbold * kbold;
  gammaB = (1.0/4.0) * waveLength * waveLength * waveLength *
    thisSphericalAbbr * kbold * kbold * kbold * kbold;
  gamma = 2 * PI * (gammaA + gammaB);
  ctf = (phaseContrast * sin(gamma)) -
    (ampContrast   * cos(gamma));
  
  return(ctf);
  
} /* calc_ctf */




