/* emUnix.h 
 * 
 * Defns for the emUnix side of the robem application.  This is
 * our transition module from the FORTRAN/VMS world into UNIX/C
 * 
 * Rob Ashmore
 * Purdue University
 * Biology dept rwa@bragg.bio.purdue.edu
*/


#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>

#include <Xm/FileSB.h>

#define PI        3.1415926535897932384626433
#define	THISTWOPI 6.283185307         

#define CUTFACE			 780
#define CUTASYM			 781
#define BULGE			 782
#define ICOSNORMAL		 783


extern unsigned int main();

extern void thetaPhi_dphshd_C (float *coremap,
			       float *origMap,
			       int ncol, int nrow, 
			       float *dphshd_map, 
			       BYTE shift, BYTE surf, 
			       int sec1, int sec2, 
			       int col1, int col2,
			       int row1, int row2,
			       int num_views,
			       int thisIrev, float surf_dens,
			       int useForMinMax,
			       float *shad_map,
			       double phi, double theta, double omega,
			       double minRadius, double maxRadius,
			       double xCen, double yCen, double zCen,
			       float radialXcen, float radialYcen, 
			       float radialZcen,
			       int lowRes, int topRadialSec, int step);

extern void emmap_rob_dphshd_C (float *map,
				float *final_map,
				float *radial_map,
				float *shad_map,
				float *origMap,
				int   numFrames,
				float surfaceDensity,
				int ncol, int nrow, 
				int firstSec, int nsec,
				int col1, int col2,
				int row1, int row2,
				char ikey,
				double thetaX, double thetaY,
				double minRadius, double maxRadius,
				double xCen, double yCen, double zCen,
				float radialXcen, float radialYcen, 
				float radialZcen,
				int lowRes, int topRadialSec);
extern void emmap_rob_fill_info ();
extern void emmap_rob_shd1 (cue_t *dphshd_map, 
			    double *tmp2d, double *tmp,
			    int ncol, int nrow,
			    int sec1, int sec2, 
			    float fract, 
			    int useForMinMax, 
			    float globAngx, float stepVal,
			    float globAngy, int whichStep,
			    float passMin, int angx, int angy, int thisFract);

extern void emmap_rob_shd2 (float *map, 
		     short *dphshd_map, 
		     int ncol, int nrow, int nsec, 
		     int sec1, int sec2, 
		     int nrec, 
		     float *Pfract, BYTE split, 
		     double *Psinx, double *Pcosx, 
		     double *Psiny, double *Pcosy,
		     int nrow_scr, int nsec_scr, 
		     BYTE dph_scratch, 
		     float *Pcenx, float *Pceny, 
		     float *Pcenz, float *Pradius);
extern void emmap_rob_dsp();
extern void mapdsp_rob_dsp ();
extern void mapdsp_rob_gplot ();
extern void dph_notilt_fast(float *coremap, 
			    int ncol, int nrow, 
			    float *dphshd_map, 
			    BYTE shift, BYTE surf, 
			    int sec1, int sec2, 
			    int num_views, 
			    int irev, float surf_dens, int useForMinMax,
			    float *shadMap);
extern void dph_nomem_tilt_fast(float *coremap, 
				float *origMap, 
				int ncol, int nrow, 
				float *dphshd_map, 
				BYTE shift, BYTE surf, 
				int sec1, int sec2, 
				int col1, int col2,
				int row1, int row2,
				int num_views, 
				int irev, float surf_dens, int useForMinMax,
				float *shadMap,
				double thetaX, double thetaY,
				double minRadius, double maxRadius,
				double xCen, double yCen, double zCen,
				float radialXcen, float radialYcen, 
				float radialZcen,
				int lowRes, int topRadialSec);
extern void dph_tilt(float *map, int ncol, int nrow, 
		     double *buf, int ndepth, int sec1,
		     double *xzslice, 
		     float *dphshd_map,
		     int irev, double theta,
		     float step_size, float surf_dens, int num_views, int nsec,
		     float *shadMap, int useForMinMax);

extern void dph_rescale_rob (int ncol, int nrow, int num_views, 
			     cue_t *cuedMap, int tryColors, 
			     int *radialStepSize);

extern int checkMemory(void *status, int line, char *file);

extern void map_1d_rob (float *map, int ncol, int nrow, int nsec, 
			float xcen, float ycen, float zcen,
			int interp, int nterp, float max_r, 
			float *rint, int *irnum, float *sum_rint,
			float *sumsq_rint, float *stdev);

extern void img_1d_rob(float *img,
		       int ncol, int nrow,
		       float xcen, float ycen,
		       int interp, int nterp,
		       float *rint, int *irnum,
		       float img_bkg);

extern void map_1d_pack_rob (BYTE *map, int ncol, int nrow, int nsec, 
			float xcen, float ycen, float zcen,
			int interp, int nterp, float max_r, 
			float *rint, int *irnum, float *sum_rint,
			float *sumsq_rint, float *stdev);

extern void img_imbed_ccf_rob(short *img1, short *img2, float *ccf, 
			      int ncol, int nrow, 
			      int xcen, int ycen, 
			      int img1_bkg, int img2_bkg, int ccf_help);
extern void map_peak_rob(float *map, 
			 int ncol, int nrow, int *dx, int *dy);

extern void img_find_peak_rob (float *r4_img, int ncol, int nrow, 
			       int *x, int *y, float *max);
extern void rob_map_roty (float *map1, float *map2, int ncol, int nrow, 
			  int firstSec, int lastSec, 
			  double rangl, float maxRadius, float minRadius);
extern void rob_map_rotx (float *map1, float *map2, int ncol, int nrow, 
			  int firstSec, int lastSec, 
			  double rangl, float maxRadius, float minRadius);
extern void rob_map_norot (float *map1, float *map2, int ncol, int nrow, 
			   int firstSec, int lastSec, 
			   float maxRadius, float minRadius,
			   double rangl);

extern void rotx(float *y,float *z,float xang);
extern void roty(float *x,float *z,float yang);
extern void rotz(float *x,float *y,float zang);

extern int askThem(char *infoStr, int systemModal, int dialogType);

double deg2rad(double degrees);
double rad2deg(double rads);

void frr(float *x, float *y, float *z) ;

extern int origNumSecs;
extern int largestDim;

extern double globTheta;
extern float radialValue[258];

extern int secondMapShade, radialMapShade, saveRadialMapShade;

extern int expStep;
extern int irev;
int img_bkg;	/* background (float) value in image */

short i2_min;	/* minimum value of short number on vax */
short i2_max;	/* maximum value of short number on vax */
int i4_min;	/* minimum value of int number on vax */
int i4_max;	/* maximum value of int number on vax */

float    map_min;	/* minimum value of map array */
float    map_max;	/* maximum value of map array */

extern	float    r4_min; /* min floating point float number on vax */
extern	float    r4_max; /* max floating point float number on vax */
extern	float    map_abang;  /* angle between rows and columns - used by  */
			     /*  emmapdsp for display purposes */
extern	float    aoverb;/* ratio of raster step size for columns vs rows */
                        /*  used by emmapdsp for scaling purposes */
float    dela;	/* pixel step size (nm) along a unit cell axis */
                /*  (i.e. spacing between columns) */
float    delb;	/* pixel step size (nm) along b unit cell axis */
                /*  (i.e. spacing between rows) */
float    delc;	/* pixel step size (nm) along c unit cell axis */
                /*  (i.e. spacing between sections) */
float    alp;		/* angle between b and c unit cell axes */
float    bet;		/* angle between a and c unit cell axes */
float    gam;		/* angle between a and b unit cell axes */
